local _, a = ...
-- local L = a.Localize
local s = SpellFlashAddon
local x = s.UpdatedVariables
local g = BittensGlobalTables
local c = g.GetTable("BittensSpellFlashLibrary")
local u = g.GetTable("BittensUtilities")

local GetComboPoints = GetComboPoints
local GetEclipseDirection = GetEclipseDirection
local GetPowerRegen = GetPowerRegen
local GetSpecialization = GetSpecialization
local GetTime = GetTime
local SPELL_POWER_ECLIPSE = SPELL_POWER_ECLIPSE
local SPELL_POWER_ENERGY = SPELL_POWER_ENERGY
local SPELL_POWER_RAGE = SPELL_POWER_RAGE
local UnitGUID = UnitGUID
local UnitPower = UnitPower
local UnitPowerMax = UnitPowerMax
local format = string.format
local min = math.min
local select = select
local tostring = tostring
local wipe = wipe

local primalPending = 0
local primalPendingFromMangle = false

local function monitorPendPrimalRage(spellID, _, _, critical)
   if critical
      and c.IdMatches(spellID, "Auto Attack", "Mangle")
      and s.Form(c.GetID("Bear Form")) then

      primalPending = GetTime()
      primalPendingFromMangle = c.IdMatches(spellID, "Mangle")
      c.Debug("Event", "Primal Fury Rage Pending")
   end
end

local function monitorConsumePrimalRage(spellID)
   if c.IdMatches(spellID, "Primal Fury Rage") then
      primalPending = 0
      c.Debug("Event", "Primal Fury Rage Happened")
   end
end

local function calcRage()
   a.Rage = c.GetPower(0, SPELL_POWER_RAGE)
   local bump = 0
   local soulMultiplier = c.HasTalent("Soul of the Forest") and GetSpecialization() == 3
         and 1.3
         or 1
   if GetTime() - primalPending < .8 then
      bump = 15
      if primalPendingFromMangle then
         bump = bump * soulMultiplier
      end
   end
   if c.IsQueued("Mangle") then
      bump = bump + 5 * soulMultiplier
   end
   a.Rage =  min(s.MaxPower("player"), a.Rage + bump)
   a.EmptyRage = s.MaxPower("player") - a.Rage
end

a.Rotations = {}

function a.PreFlash()
   a.CatForm = s.Form(c.GetID("Cat Form"))
   a.BearForm = s.Form(c.GetID("Bear Form"))
end


----------------------------------------------------------------------- Balance
local pendingSecondDot = 0

a.MoonfireTick = 2
a.SunfireTick = 2

local updateMoonfire = function()
   a.MoonfireTick = c.GetHastedTime(2)
   c.Debug("Event", "Moonfire ticks every", a.MoonfireTick)
end

local updateSunfire = function()
   a.SunfireTick = c.GetHastedTime(2)
   c.Debug("Event", "Sunfire ticks every", a.SunfireTick)
end

a.Rotations.Balance = {
   Spec = 1,

   UsefulStats = {
      "Intellect", "Spell Hit", "Hit from Spirit", "Crit", "Haste"
   },

   NoFlashInCombat = function()
      a.Energy = UnitPower("player", SPELL_POWER_ECLIPSE)
      a.Alignment = c.GetBuffDuration(
         "Celestial Alignment", false, false, true)

      if a.Alignment > 0 then
         a.Solar = true
         a.Lunar = true
      else
         a.Solar = c.HasBuff("Eclipse (Solar)")
         a.Lunar = c.HasBuff("Eclipse (Lunar)")
      end
      a.GoingUp = GetEclipseDirection() ~= "moon"

      a.EclipsePending = false
      if a.Alignment == 0 then
         local bump = 0
         if c.IsCasting("Starfire") then
            if a.GoingUp then
               bump = 20
            end
         elseif c.IsCasting("Wrath") then
            if not a.GoingUp then
               bump = -15
            end
         elseif c.IsCasting("Starsurge") then
            if a.GoingUp then
               bump = 20
            else
               bump = -20
            end
         end
         -- @todo danielp 2014-11-11: astral insight no longer exists
         -- if c.IsCasting("Astral Communion")
         --    and c.HasBuff("Astral Insight") then
         -- 
         --    if a.GoingUp then
         --       bump = bump + 100
         --    else
         --       bump = bump - 100
         --    end
         -- end
         if bump ~= 0 then
            -- changed in 6.0, I think???
            -- if c.HasSpell("Euphoria")
            --    and not a.Solar
            --    and not a.Lunar then
            -- 
            --    bump = bump * 2
            -- end
            a.Energy = a.Energy + bump
            if a.Energy <= -100 then
               a.GoingUp = true
               a.Lunar = true
               a.EclipsePending = true
            elseif a.Energy >= 100 then
               a.GoingUp = false
               a.Solar = true
               a.EclipsePending = true
            elseif a.Energy >= 0 then
               a.Lunar = false
            elseif a.Energy <= 0 then
               a.Solar = false
            end
         end
      end

      local now = GetTime()
      if now - pendingSecondDot < .8
         or (a.Alignment > 0 and c.IsCasting("Moonfire", "Sunfire")) then

         a.Moonfire = 14
         a.Sunfire = 14
      else
         a.Moonfire = c.GetMyDebuffDuration("Moonfire", false, false, true)
         a.Sunfire = c.GetMyDebuffDuration("Sunfire", false, false, true)
      end

      -- if now - pendingLunarShower < 0.8
      --    or c.IsCasting("Moonfire", "Sunfire") then
      -- 
      --    a.LunarShower = 3 - c.GetBusyTime()
      -- else
      --    a.LunarShower = c.GetBuffDuration("Lunar Shower")
      -- end

      c.FlashAll(
         "Starfall",
         "Force of Nature: Balance",
         "Incarnation: Chosen of Elune",
         "Nature's Vigil",
         "Celestial Alignment",
         "Renewal",
         "Soothe",
         "Solar Beam")

      c.PriorityFlash(
         "Astral Communion Instant",
         "Starsurge under Shooting Stars",
         "Moonfire under Eclipse",
         "Sunfire under Eclipse",
         "Moonfire",
         "Sunfire",
         "Starsurge",
         "Starfire under Celestial Alignment",
         "Wrath under Celestial Alignment",
         "Starfire",
         "Wrath")
   end,

   NoFlashOutOfCombat = function()
      c.FlashAll("Healing Touch Solo")
   end,

   FlashAlways = function()
      c.FlashAll("Mark of the Wild", "Moonkin Form")
   end,

   NoCastSucceeded = function(info)
      if c.InfoMatches(info, "Moonfire", "Sunfire") then
         if c.HasBuff("Celestial Alignment", true, false, true) then
            pendingSecondDot = GetTime()
            c.Debug("Event", "Pending Second DoT")
         end
         -- pendingLunarShower = GetTime()
         -- c.Debug("Event", "Lunar Shower Pending")
      end
   end,

   NoAuraApplied = function(spellID)
      if c.IdMatches(spellID, "Moonfire") then
         updateMoonfire()
      elseif c.IdMatches(spellID, "Starfire") then
         updateSunfire()
      -- elseif c.IdMatches(spellID, "Lunar Shower") then
      --    pendingLunarShower = 0
      --    c.Debug("Event", "Lunar Shower Happened")
      end
   end,

   NoSpellDamage = function(spellID, _, _, crit)
      if crit then
         if c.IdMatches(spellID, "Starsurge") then
            updateMoonfire()
            updateSunfire()
         elseif c.IdMatches(spellID, "Wrath") then
            updateSunfire()
         elseif c.IdMatches(spellID, "Starfire") then
            updateMoonfire()
         end
      end
   end,

   NoExtraDebugInfo = function()
      return format(
         "e:%d s:%s l:%s u:%s p:%s m:%.1f s:%.1f l:%.1f a:%.1f",
         a.Energy,
         tostring(a.Solar),
         tostring(a.Lunar),
         tostring(a.GoingUp),
         tostring(a.EclipsePending),
         a.Moonfire,
         a.Sunfire,
         a.LunarShower,
         a.Alignment)
   end,
}

------------------------------------------------------------------------- Feral
local function commonBleed()
   return 1
      * (c.HasBuff("Savage Roar") and 1.4 or 1)
      * (c.HasBuff("Tiger's Fury") and 1.15 or 1)
      * (c.HasBuff("Bloodtalons") and 1.3 or 1)
end

-- tracking the damage multipliers for targets
local rip = {}
local function registerRipBleed(id, cp)
   local data = u.GetOrMakeTable(rip, id)
   data.cp = cp
   data.strength = commonBleed()
   c.Debug("Event", format("Rip %d@%d on %s", data.cp, data.strength, id))
end

local function removeRipBleed(id)
   local data = rip[id]
   if data then
      data.cp = 0
      data.strength = 0
   end
end

function a.ripWouldBeStronger(cp)
   local guid = UnitGUID("target")
   if not guid then
      return false
   end

   local current = rip[guid]
   if not current then
      return false
   end

   -- less combo points == weaker damage, so no overwrite.
   if current.cp < cp then
      return false
   end

   local strength = commonBleed()

   return strength > current.strength
      or (strength == current.strength and a.Rip <= 7.2)
end


local rake = {}
local function registerRakeBleed(id)
   local stealth = c.HasSpell("Improved Rake")
      and (c.HasBuff("Prowl")
              or c.HasBuff("Shadowmeld")
              or c.HasBuff("Incarnation: King of the Jungle"))

   local strength = commonBleed() * (stealth and 2 or 1)

   if id then
      rake[id] = strength
      c.Debug("Event", format("Rake @%d on %s", strength, id))
   end

   return strength
end

local function removeRakeBleed(id)
   rake[id] = 0
end

function a.rakeWouldBeStronger(bonus)
   local guid = UnitGUID("target")
   if not guid then return false end

   local current = rake[guid] or 0
   local strength = registerRakeBleed() * (bonus or 1)

   return strength > current
      or (strength == current and a.Rake <= 4.5)
end

a.Rotations.Feral = {
   Spec = 2,

   UsefulStats = { "Agility", "Multistrike", "Crit", "Haste", "Versatility" },

   PreFlash = function()
      a.Regen = select(2, GetPowerRegen())
      a.Energy = UnitPower("player", SPELL_POWER_ENERGY)
      a.MaxEnergy = UnitPowerMax("player", SPELL_POWER_ENERGY)
      a.MissingEnergy = a.MaxEnergy - a.Energy

      -- @todo danielp 2015-01-24: GetUnitPower("player", 4) ??
      a.ComboPoints = GetComboPoints("player", "target")

      a.Rake = c.GetMyDebuffDuration("Rake Bleed")
      a.Rip = c.GetMyDebuffDuration("Rip")
      a.SavageRoar = c.GetBuffDuration("Savage Roar")

      a.InExecute = c.GetHealthPercent("target") <= 25
   end,

   FlashInCombat = function()
      c.FlashAll(
         "Force of Nature: Feral",
         "Renewal",
         "Soothe",
         "Skull Bash",
         "Survival Instincts under 30"
      )

      -- if you ain't in it, we can't do much more. :)
      if c.FlashAll("Cat Form") then
         return
      end

      c.DelayPriorityFlash(
         "Wild Charge: Cat",
         -- displacer_beast,if=movement.distance>10
         -- dash,if=movement.distance&buff.displacer_beast.down&buff.wild_charge_movement.down
         "Rake with stealth",
         "Tiger's Fury",
         "Incarnation: King of the Jungle",
         "Berserk for Feral",
         "Shadowmeld for Feral",
         "Ferocious Bite to extend Rip",
         "Healing Touch for Bloodtalons",
         "Savage Roar to refresh",
         "Thrash (Cat Form)",
         "Thrash (Cat Form) AoE delay",

         -- finisher
         "Rip",
         "Savage Roar",
         "Thrash (Cat Form) at 5",
         "Ferocious Bite at 5",

         -- maintain
         "Rake",
         "Moonfire for Feral",
         "Rake Late",

         -- generate
         "Swipe",
         "Shred"
      )
   end,

   FlashOutOfCombat = function()
      c.FlashAll(
         "Healing Touch Solo"
      )

      if x.EnemyDetected then
         c.DelayPriorityFlash(
            "Healing Touch for Bloodtalons pre-pull",
            "Cat Form",
            "Prowl",
            "Incarnation: King of the Jungle pre-pull",
            "Wild Charge: Cat",
            "Rake with stealth",
            "Swipe",
            "Shred"
         )
      end
   end,

   FlashAlways = function()
      c.FlashAll("Mark of the Wild")
   end,

   LeftCombat = function()
      wipe(rip)
      wipe(rake)
   end,

   CastSucceeded_FromLog = function(spellID, _, targetID)
      if c.IdMatches(spellID, "Rip") then
         -- 4 is the combo points magic number
         registerRipBleed(targetID, UnitPower("player", 4))
      elseif c.IdMatches(spellID, "Rake") then
         registerRakeBleed(targetID)
      end
   end,

   AuraRemoved = function(spellID, _, targetID)
      if c.IdMatches(spellID, "Rip") then
         removeRipBleed(targetID)
      elseif c.IdMatches(spellID, "Rake") then
         removeRakeBleed(targetID)
      end
   end,

   -- ExtraDebugInfo = function()
   --    return format(
   --       "b:%.1f r:%.1f e:%.1f c:%d c:%s s:%.1f d:%d b:%.1f t:%.1f k:%.1f r:%.1f r:%.1f x:%s",
   --       c.GetBusyTime(),
   --       a.Rage,
   --       a.Energy,
   --       a.CP,
   --       tostring(not not a.Clearcasting),
   --       a.Swiftness,
   --       a.DreamStacks,
   --       a.Berserk,
   --       a.TigersFury,
   --       a.King,
   --       a.Rip,
   --       a.Roar,
   --       tostring(not not a.InExecute))
   -- end,
}

---------------------------------------------------------------------- Guardian
a.Rotations.Guardian = {
   Spec = 3,
   -- AoEColor = "aqua",

   UsefulStats = {
      "Stamina", "Bonus Armor", "Mastery", "Versatility", "Multistrike", "Haste", "Crit"
   },

   FlashInCombat = function()
      calcRage()

      c.FlashMitigationBuffs(
         1,
         "Tooth and Claw",
         c.COMMON_TANKING_BUFFS,
         "Healing Touch Mitigation Delay",
         "Cenarion Ward for Guardian",
         "Bristling Fur",
         "Barkskin",
         "Renewal for Guardian",
         "Survival Instincts Glyphed",
         "Incarnation: Son of Ursoc",
         "Survival Instincts Unglyphed"
      )

      c.FlashAll(
         "Frenzied Regeneration",
         "Savage Defense",
         "Maul for Guardian",
         "Skull Bash",
         "Growl"
      )

      c.DelayPriorityFlash(
         "Berserk for Guardian",
         "Heart of the Wild: Guardian",
         "Rejuvenation for Guardian",
         "Nature's Vigil",
         "Healing Touch for Guardian",
         "Pulverize",
         "Lacerate for Pulverize",
         -- incarnation
         "Lacerate for Debuff",
         "Thrash (Bear Form) if down",
         "Mangle",
         "Thrash (Bear Form) early",
         "Lacerate"
      )
   end,

   FlashOutOfCombat = function()
      c.FlashAll("Healing Touch Solo")
   end,

   FlashAlways = function()
      c.FlashAll("Mark of the Wild", "Bear Form")
   end,

   SpellDamage = monitorPendPrimalRage,

   AutoAttack = monitorPendPrimalRage,

   Energized = monitorConsumePrimalRage,

   ExtraDebugInfo = function()
      return format("r:%.1f b:%.2f", a.Rage, c.GetBusyTime())
   end,
}

------------------------------------------------------------------- Restoration
a.Rotations.Restoration = {
   Spec = 4,

   UsefulStats = { "Intellect", "Spirit", "Crit", "Haste" },

   FlashInCombat = function()
      c.FlashAll(
         "Force of Nature: Restoration",
         "Lifebloom",
         "Healing Touch for Restoration",
         "Regrowth",
         "Wild Mushroom: Restoration",
         "Renewal",
         "Cenarion Ward for Restoration",
         "Soothe")
   end,

   FlashAlways = function()
      c.FlashAll("Mark of the Wild")
   end,

   AuraApplied = function(spellID, target)
      if c.IdMatches(spellID, "Lifebloom") then
         a.LifebloomTarget = target
         c.Debug("Event", "Lifebloom on", target)
      end
   end,
}
